@@ -4,13 +4,40 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
+/*********************************************************************/
+/*                       MALLOC/FREE WRAPPERS                        */
+/*********************************************************************/
+#define BLOCK_SIGALRM_START          \
+{                                    \
+  int block=PARI_SIGINT_block;       \
+  PARI_SIGINT_block = 2;             \
+  MT_SIGINT_BLOCK(block);
+
+#define BLOCK_SIGINT_START           \
+{                                    \
+  int block=PARI_SIGINT_block;       \
+  PARI_SIGINT_block = 1;             \
+  MT_SIGINT_BLOCK(block);
+
+#define BLOCK_SIGINT_END             \
+  PARI_SIGINT_block = block;         \
+  MT_SIGINT_UNBLOCK(block);          \
+  if (!block && PARI_SIGINT_pending) \
+  {                                  \
+    int sig = PARI_SIGINT_pending;   \
+    PARI_SIGINT_pending = 0;         \
+    raise(sig);                      \
+  }                                  \
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                          CONSTRUCTORS                           */
@@ -149,7 +176,7 @@ mkpolmod(GEN x, GEN y) { retmkpolmod(x,y); }
 INLINE GEN
 mkfrac(GEN x, GEN y) { retmkfrac(x,y); }
 INLINE GEN
-mkfracss(long x, long y) { retmkfrac(stoi(x),stoi(y)); }
+mkfracss(long x, long y) { retmkfrac(stoi(x),utoipos(y)); }
 /* q = n/d a t_FRAC or t_INT; recover (n,d) */
 INLINE void
 Qtoss(GEN q, long *n, long *d)
@@ -308,6 +335,15 @@ pol_xnall(long n, long v)
   if (n < 0) retmkrfrac(gen_1, pol_xn(-n,v));
   return pol_xn(n, v);
 }
+/* x^n, assume n >= 0 */
+INLINE GEN
+polxn_Flx(long n, long sv) {
+  long i, a = n+2;
+  GEN p = cgetg(a+1, t_VECSMALL);
+  p[1] = sv;
+  for (i = 2; i < a; i++) p[i] = 0;
+  p[a] = 1; return p;
+}
 INLINE GEN
 pol_1(long v) {
   GEN p = cgetg(3, t_POL);
@@ -538,6 +574,32 @@ vecsmalltrunc_init(long l)
 INLINE void
 vecsmalltrunc_append(GEN x, long t) { x[ lg(x) ] = t; lg_increase(x); }
 
+/*******************************************************************/
+/*                                                                 */
+/*                    STRING HASH FUNCTIONS                        */
+/*                                                                 */
+/*******************************************************************/
+INLINE ulong
+hash_str(const char *str)
+{
+  ulong hash = 5381UL, c;
+  while ( (c = (ulong)*str++) )
+    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  return hash;
+}
+INLINE ulong
+hash_str_len(const char *str, long len)
+{
+  ulong hash = 5381UL;
+  long i;
+  for (i = 0; i < len; i++)
+  {
+    ulong c = (ulong)*str++;
+    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  }
+  return hash;
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                        VEC / COL / VECSMALL                     */
@@ -822,6 +884,16 @@ RgV_is_QV(GEN x)
     if (!is_rational_t(typ(gel(x,i)))) return 0;
   return 1;
 }
+INLINE long
+RgV_isin_i(GEN v, GEN x, long n)
+{
+  long i;
+  for (i = 1; i <= n; i++)
+    if (gequal(gel(v,i), x)) return i;
+  return 0;
+}
+INLINE long
+RgV_isin(GEN v, GEN x) { return RgV_isin_i(v, x, lg(v)-1); }
 
 /********************************************************************/
 /**                                                                **/
@@ -853,7 +925,7 @@ pari_stack_alloc(pari_stack *s, long nb)
   {
     while (s->n+nb > alloc) alloc <<= 1;
   }
-  *sdat = pari_realloc(*sdat,alloc*s->size);
+  pari_realloc_ip(sdat,alloc*s->size);
   s->alloc = alloc;
 }
 
@@ -1004,15 +1076,25 @@ rowpermute(GEN x, GEN p)
 /*                          PERMUTATIONS                           */
 /*                                                                 */
 /*******************************************************************/
+INLINE GEN
+identity_zv(long n)
+{
+  GEN v = cgetg(n+1, t_VECSMALL);
+  long i;
+  for (i = 1; i <= n; i++) v[i] = i;
+  return v;
+}
+INLINE GEN
+identity_ZV(long n)
+{
+  GEN v = cgetg(n+1, t_VEC);
+  long i;
+  for (i = 1; i <= n; i++) gel(v,i) = utoipos(i);
+  return v;
+}
 /* identity permutation */
 INLINE GEN
-identity_perm(long n)
-{
-  GEN perm = cgetg(n+1, t_VECSMALL);
-  long i;
-  for (i = 1; i <= n; i++) perm[i] = i;
-  return perm;
-}
+identity_perm(long n) { return identity_zv(n); }
 
 /* assume d <= n */
 INLINE GEN
@@ -1030,6 +1112,10 @@ INLINE GEN
 perm_mul(GEN s, GEN x)
 { pari_APPLY_long(s[x[i]]) }
 
+INLINE GEN
+perm_sqr(GEN x)
+{ pari_APPLY_long(x[x[i]]) }
+
 /* Compute the inverse (reciprocal) of a permutation. */
 INLINE GEN
 perm_inv(GEN x)
@@ -1049,32 +1135,6 @@ perm_conj(GEN s, GEN t)
   return v;
 }
 
-/*********************************************************************/
-/*                       MALLOC/FREE WRAPPERS                        */
-/*********************************************************************/
-#define BLOCK_SIGALRM_START          \
-{                                    \
-  int block=PARI_SIGINT_block;       \
-  PARI_SIGINT_block = 2;             \
-  MT_SIGINT_BLOCK(block);
-
-#define BLOCK_SIGINT_START           \
-{                                    \
-  int block=PARI_SIGINT_block;       \
-  PARI_SIGINT_block = 1;             \
-  MT_SIGINT_BLOCK(block);
-
-#define BLOCK_SIGINT_END             \
-  PARI_SIGINT_block = block;         \
-  MT_SIGINT_UNBLOCK(block);          \
-  if (!block && PARI_SIGINT_pending) \
-  {                                  \
-    int sig = PARI_SIGINT_pending;   \
-    PARI_SIGINT_pending = 0;         \
-    raise(sig);                      \
-  }                                  \
-}
-
 INLINE void
 pari_free(void *pointer)
 {
@@ -1108,6 +1168,18 @@ pari_realloc(void *pointer, size_t size)
   if (!tmp) pari_err(e_MEM);
   return tmp;
 }
+INLINE void
+pari_realloc_ip(void **pointer, size_t size)
+{
+  char *tmp;
+  BLOCK_SIGINT_START;
+  if (!*pointer) tmp = (char *) malloc(size);
+  else tmp = (char *) realloc(*pointer,size);
+  if (!tmp) pari_err(e_MEM);
+  *pointer = tmp;
+  BLOCK_SIGINT_END;
+}
+
 INLINE void*
 pari_calloc(size_t size)
 {
@@ -1116,9 +1188,10 @@ pari_calloc(size_t size)
 }
 INLINE GEN
 cgetalloc(long t, size_t l)
-{
+{ /* evallg may raise e_OVERFLOW, which would leak x */
+  ulong x0 = evaltyp(t) | evallg(l);
   GEN x = (GEN)pari_malloc(l * sizeof(long));
-  x[0] = evaltyp(t) | evallg(l); return x;
+  x[0] = x0; return x;
 }
 
 /*******************************************************************/
@@ -1126,7 +1199,7 @@ cgetalloc(long t, size_t l)
 /*                       GARBAGE COLLECTION                        */
 /*                                                                 */
 /*******************************************************************/
-/* copy integer x as if we had avma = av */
+/* copy integer x as if we had set_avma(av) */
 INLINE GEN
 icopy_avma(GEN x, pari_sp av)
 {
@@ -1136,7 +1209,7 @@ icopy_avma(GEN x, pari_sp av)
   y[0] = evaltyp(t_INT)|evallg(lx);
   return y;
 }
-/* copy leaf x as if we had avma = av */
+/* copy leaf x as if we had set_avma(av) */
 INLINE GEN
 leafcopy_avma(GEN x, pari_sp av)
 {
@@ -1152,24 +1225,24 @@ gerepileuptoleaf(pari_sp av, GEN x)
   long lx;
   GEN q;
 
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
   lx = lg(x);
   q = ((GEN)av) - lx;
-  avma = (pari_sp)q;
+  set_avma((pari_sp)q);
   while (--lx >= 0) q[lx] = x[lx];
   return q;
 }
 INLINE GEN
 gerepileuptoint(pari_sp av, GEN x)
 {
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
-  avma = (pari_sp)icopy_avma(x, av);
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
+  set_avma((pari_sp)icopy_avma(x, av));
   return (GEN)avma;
 }
 INLINE GEN
 gerepileupto(pari_sp av, GEN x)
 {
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
   switch(typ(x))
   { /* non-default = !is_recursive_t(tq) */
     case t_INT: return gerepileuptoint(av, x);
@@ -1189,21 +1262,26 @@ gerepilecopy(pari_sp av, GEN x)
   if (is_recursive_t(typ(x)))
   {
     GENbin *p = copy_bin(x);
-    avma = av; return bin_copy(p);
+    set_avma(av); return bin_copy(p);
   }
   else
   {
-    avma = av;
+    set_avma(av);
     if (x < (GEN)av) {
       if (x < (GEN)pari_mainstack->bot) new_chunk(lg(x));
       x = leafcopy_avma(x, av);
-      avma = (pari_sp)x;
+      set_avma((pari_sp)x);
     } else
       x = leafcopy(x);
     return x;
   }
 }
 
+INLINE void
+guncloneNULL(GEN x) { if (x) gunclone(x); }
+INLINE void
+guncloneNULL_deep(GEN x) { if (x) gunclone_deep(x); }
+
 /* Takes an array of pointers to GENs, of length n. Copies all
  * objects to contiguous locations and cleans up the stack between
  * av and avma. */
@@ -1212,7 +1290,7 @@ gerepilemany(pari_sp av, GEN* gptr[], int n)
 {
   int i;
   for (i=0; i<n; i++) *gptr[i] = (GEN)copy_bin(*gptr[i]);
-  avma = av;
+  set_avma(av);
   for (i=0; i<n; i++) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
 }
 
@@ -1226,7 +1304,7 @@ gerepileall(pari_sp av, int n, ...)
     GEN *gptr[10];
     for (i=0; i<n; i++)
     { gptr[i] = va_arg(a,GEN*); *gptr[i] = (GEN)copy_bin(*gptr[i]); }
-    avma = av;
+    set_avma(av);
     for (--i; i>=0; i--) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
 
   }
@@ -1235,7 +1313,7 @@ gerepileall(pari_sp av, int n, ...)
     GEN **gptr = (GEN**)  pari_malloc(n*sizeof(GEN*));
     for (i=0; i<n; i++)
     { gptr[i] = va_arg(a,GEN*); *gptr[i] = (GEN)copy_bin(*gptr[i]); }
-    avma = av;
+    set_avma(av);
     for (--i; i>=0; i--) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
     pari_free(gptr);
   }
@@ -1247,7 +1325,7 @@ gerepilecoeffs(pari_sp av, GEN x, int n)
 {
   int i;
   for (i=0; i<n; i++) gel(x,i) = (GEN)copy_bin(gel(x,i));
-  avma = av;
+  set_avma(av);
   for (i=0; i<n; i++) gel(x,i) = bin_copy((GENbin*)x[i]);
 }
 
@@ -1274,7 +1352,7 @@ INLINE void
 cgiv(GEN x)
 {
   pari_sp av = (pari_sp)(x+lg(x));
-  if (isonstack((GEN)av)) avma = av;
+  if (isonstack((GEN)av)) set_avma(av);
 }
 
 INLINE void
@@ -1316,10 +1394,24 @@ gtodouble(GEN x)
     pari_sp av = avma;
     x = gtofp(x, DEFAULTPREC);
     if (typ(x)!=t_REAL) pari_err_TYPE("gtodouble [t_REAL expected]", x);
-    avma = av;
+    set_avma(av);
   }
   return rtodbl(x);
 }
+
+INLINE int
+gisdouble(GEN x, double *g)
+{
+  if (typ(x)!=t_REAL) {
+    pari_sp av = avma;
+    x = gtofp(x, DEFAULTPREC);
+    if (typ(x)!=t_REAL) pari_err_TYPE("gisdouble [t_REAL expected]", x);
+    set_avma(av);
+  }
+  if (expo(x) >= 0x3ff) return 0;
+  *g = rtodbl(x); return 1;
+}
+
 INLINE long
 gtos(GEN x) {
   if (typ(x) != t_INT) pari_err_TYPE("gtos [integer expected]",x);
@@ -1452,7 +1544,7 @@ affgr(GEN x, GEN y)
     case t_INT:  affir(x,y); break;
     case t_REAL: affrr(x,y); break;
     case t_FRAC: rdiviiz(gel(x,1),gel(x,2), y); break;
-    case t_QUAD: av = avma; affgr(quadtofp(x,realprec(y)), y); avma = av; break;
+    case t_QUAD: av = avma; affgr(quadtofp(x,realprec(y)), y); set_avma(av); break;
     default: pari_err_TYPE2("=",x,y);
   }
 }
@@ -1467,7 +1559,7 @@ affc_fixlg(GEN x, GEN res)
   }
   else
   {
-    avma = (pari_sp)(res+3);
+    set_avma((pari_sp)(res+3));
     res = cgetr(realprec(gel(res,1)));
     affrr_fixlg(x, res);
   }
@@ -1556,8 +1648,8 @@ Fp_add(GEN a, GEN b, GEN m)
   {
     GEN t = subii(p, m);
     s = signe(t);
-    if (!s) { avma = av; return gen_0; }
-    if (s < 0) { avma = (pari_sp)p; return p; }
+    if (!s) return gc_const(av, gen_0);
+    if (s < 0) return gc_const((pari_sp)p, p);
     if (cmpii(t, m) < 0) return gerepileuptoint(av, t); /* general case ! */
     p = remii(t, m);
   }
@@ -1580,7 +1672,7 @@ Fp_sub(GEN a, GEN b, GEN m)
   else
   {
     GEN t = addii(p, m);
-    if (!s) { avma = av; return gen_0; }
+    if (!s) return gc_const(av, gen_0);
     if (s > 0) return gerepileuptoint(av, t); /* general case ! */
     p = modii(t, m);
   }
@@ -1637,7 +1729,7 @@ Fp_mul(GEN a, GEN b, GEN m)
   GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk(lg(a)+lg(b)+(lg(m)<<1));
   p = mulii(a,b);
-  avma = av; return modii(p,m);
+  set_avma(av); return modii(p,m);
 }
 INLINE GEN
 Fp_sqr(GEN a, GEN m)
@@ -1646,7 +1738,7 @@ Fp_sqr(GEN a, GEN m)
   GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk((lg(a)+lg(m))<<1);
   p = sqri(a);
-  avma = av; return remii(p,m); /*Use remii: p >= 0 */
+  set_avma(av); return remii(p,m); /*Use remii: p >= 0 */
 }
 INLINE GEN
 Fp_mulu(GEN a, ulong b, GEN m)
@@ -1661,7 +1753,7 @@ Fp_mulu(GEN a, ulong b, GEN m)
     GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
     (void)new_chunk(lg(a)+1+(l<<1));
     p = muliu(a,b);
-    avma = av; return modii(p,m);
+    set_avma(av); return modii(p,m);
   }
 }
 INLINE GEN
@@ -1683,7 +1775,7 @@ Fp_muls(GEN a, long b, GEN m)
     GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
     (void)new_chunk(lg(a)+1+(l<<1));
     p = mulis(a,b);
-    avma = av; return modii(p,m);
+    set_avma(av); return modii(p,m);
   }
 }
 
@@ -1704,11 +1796,32 @@ Fp_invsafe(GEN a, GEN m)
 INLINE GEN
 Fp_div(GEN a, GEN b, GEN m)
 {
-  pari_sp av=avma;
-  GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
+  pari_sp av = avma;
+  GEN p;
+  if (lgefint(b) == 3)
+  {
+    a = Fp_divu(a, b[2], m);
+    if (signe(b) < 0) a = Fp_neg(a, m);
+    return a;
+  }
+  /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk(lg(a)+(lg(m)<<1));
   p = mulii(a, Fp_inv(b,m));
-  avma = av; return modii(p,m);
+  set_avma(av); return modii(p,m);
+}
+INLINE GEN
+Fp_divu(GEN x, ulong a, GEN p)
+{
+  pari_sp av = avma;
+  ulong b;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2], xp = umodiu(x, pp);
+    return xp? utoipos(Fl_div(xp, a % pp, pp)): gen_0;
+  }
+  x = Fp_red(x, p);
+  b = Fl_neg(Fl_div(umodiu(x,a), umodiu(p,a), a), a); /* x + pb = 0 (mod a) */
+  return gerepileuptoint(av, diviuexact(addmuliu(x, p, b), a));
 }
 
 INLINE GEN
@@ -1786,7 +1899,7 @@ submulii(GEN x, GEN y, GEN z)
   lz = lgefint(z);
   av = avma; (void)new_chunk(lx+ly+lz); /* HACK */
   t = mulii(z, y);
-  avma = av; return subii(x,t);
+  set_avma(av); return subii(x,t);
 }
 /* y*z - x */
 INLINE GEN
@@ -1801,7 +1914,7 @@ mulsubii(GEN y, GEN z, GEN x)
   lz = lgefint(z);
   av = avma; (void)new_chunk(lx+ly+lz); /* HACK */
   t = mulii(z, y);
-  avma = av; return subii(t,x);
+  set_avma(av); return subii(t,x);
 }
 
 /* x - u*y */
@@ -1814,7 +1927,7 @@ submuliu(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return subii(x, y);
+  set_avma(av); return subii(x, y);
 }
 /* x + u*y */
 INLINE GEN
@@ -1826,7 +1939,7 @@ addmuliu(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return addii(x, y);
+  set_avma(av); return addii(x, y);
 }
 /* x - u*y */
 INLINE GEN
@@ -1838,7 +1951,7 @@ submuliu_inplace(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return subii(x, y);
+  set_avma(av); return subii(x, y);
 }
 /* x + u*y */
 INLINE GEN
@@ -1850,7 +1963,7 @@ addmuliu_inplace(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return addii(x, y);
+  set_avma(av); return addii(x, y);
 }
 /* ux + vy */
 INLINE GEN
@@ -1865,7 +1978,7 @@ lincombii(GEN u, GEN v, GEN x, GEN y)
   av = avma; (void)new_chunk(lx+ly+lgefint(u)+lgefint(v)); /* HACK */
   p1 = mulii(u,x);
   p2 = mulii(v,y);
-  avma = av; return addii(p1,p2);
+  set_avma(av); return addii(p1,p2);
 }
 
 /*******************************************************************/
@@ -1887,6 +2000,8 @@ is_noncalc_t(long tx) { return (tx) >= t_LIST; }
 INLINE int
 is_rational_t(long t) { return (t == t_INT || t == t_FRAC); }
 INLINE int
+is_qfb_t(long t) { return (t == t_QFI || t == t_QFR); }
+INLINE int
 is_real_t(long t) { return (t == t_INT || t == t_REAL || t == t_FRAC); }
 INLINE int
 is_recursive_t(long t) { return lontyp[t]; }
@@ -1986,6 +2101,13 @@ mul_content(GEN cx, GEN cy)
 INLINE GEN
 inv_content(GEN c) { return c? ginv(c): NULL; }
 INLINE GEN
+div_content(GEN cx, GEN cy)
+{
+  if (!cy) return cx;
+  if (!cx) return ginv(cy);
+  return gdiv(cx,cy);
+}
+INLINE GEN
 mul_denom(GEN dx, GEN dy)
 {
   if (!dx) return dy;
@@ -2113,11 +2235,8 @@ INLINE long
 sturm(GEN x) { return sturmpart(x, NULL, NULL); }
 
 INLINE long
-gval(GEN x, long v) {
-  pari_sp av = avma;
-  long n = gvaluation(x, pol_x(v));
-  avma = av; return n;
-}
+gval(GEN x, long v)
+{ pari_sp av = avma; return gc_long(av, gvaluation(x, pol_x(v))); }
 
 INLINE void
 RgX_shift_inplace_init(long v)
@@ -2127,18 +2246,20 @@ INLINE GEN
 RgX_shift_inplace(GEN x, long v)
 {
   long i, lx;
-  GEN y, z;
+  GEN z;
   if (!v) return x;
   lx = lg(x);
   if (lx == 2) return x;
-  y = x + v;
   z = x + lx;
-  /* stackdummy from normalizepol: move it up */
-  if (lg(z) != v) x[lx + v] = z[0];
-  for (i = lx-1; i >= 2; i--) gel(y,i) = gel(x,i);
-  for (i = v+1;  i >= 2; i--) gel(x,i) = gen_0;
-  /* leave x[1] alone: it is correct */
-  x[0] = evaltyp(t_POL) | evallg(lx+v); return x;
+  /* stackdummy's from normalizepol */
+  while (lg(z) != v) z += lg(z);
+  z += v;
+  for (i = lx-1; i >= 2; i--) gel(--z,0) = gel(x,i);
+  for (i = 0;  i < v; i++) gel(--z,0) = gen_0;
+  z -= 2;
+  z[1] = x[1];
+  z[0] = evaltyp(t_POL) | evallg(lx+v);
+  stackdummy((pari_sp)z, (pari_sp)x); return z;
 }
 
 
@@ -2227,6 +2348,10 @@ FlxqX_div(GEN x, GEN y, GEN T, ulong p) { return FlxqX_divrem(x, y, T, p, NULL);
 INLINE GEN
 F2xqX_div(GEN x, GEN y, GEN T) { return F2xqX_divrem(x, y, T, NULL); }
 
+INLINE GEN
+FpXY_Fq_evaly(GEN Q, GEN y, GEN T, GEN p, long vx)
+{ return T ? FpXY_FpXQ_evaly(Q, y, T, p, vx): FpXY_evaly(Q, y, p, vx); }
+
 /* FqX */
 INLINE GEN
 FqX_red(GEN z, GEN T, GEN p) { return T? FpXQX_red(z, T, p): FpX_red(z, p); }
@@ -2335,6 +2460,9 @@ FqXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
 
 /*FqXn*/
 INLINE GEN
+FqXn_expint(GEN x, long n, GEN T, GEN p)
+{ return T? FpXQXn_expint(x,n,T,p): FpXn_expint(x,n,p); }
+INLINE GEN
 FqXn_exp(GEN x, long n, GEN T, GEN p)
 { return T? FpXQXn_exp(x,n,T,p): FpXn_exp(x,n,p); }
 INLINE GEN
@@ -2447,6 +2575,11 @@ gsubgs(GEN y, long s) { return gaddgs(y, -s); }
 INLINE GEN
 gdivsg(long s, GEN y) { return gdiv(stoi(s), y); }
 
+INLINE GEN
+gmax_shallow(GEN x, GEN y) { return gcmp(x,y)<0? y: x; }
+INLINE GEN
+gmin_shallow(GEN x, GEN y) { return gcmp(x,y)<0? x: y; }
+
 /* x t_COMPLEX */
 INLINE GEN
 cxnorm(GEN x) { return gadd(gsqr(gel(x,1)), gsqr(gel(x,2))); }
@@ -2471,13 +2604,9 @@ quadnorm(GEN q)
 INLINE GEN
 quad_disc(GEN x)
 {
-  GEN Q = gel(x,1), b = gel(Q,3), c = gel(Q,2), c4;
-  if (is_pm1(b))
-  {
-    pari_sp av = avma; (void)new_chunk(lgefint(c) + 1);
-    c4 = shifti(c,2); avma = av; return subsi(1, c4);
-  }
-  c4 = shifti(c,2); togglesign_safe(&c4); return c4;
+  GEN Q = gel(x,1), b = gel(Q,3), c = gel(Q,2), c4 = shifti(c,2);
+  if (is_pm1(b)) return subsi(1, c4);
+  togglesign_safe(&c4); return c4;
 }
 INLINE GEN
 qfb_disc3(GEN x, GEN y, GEN z) { return subii(sqri(y), shifti(mulii(x,z),2)); }
@@ -2530,33 +2659,33 @@ powIs(long n) {
 /*                                                                 */
 /*******************************************************************/
 INLINE void mpexpz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpexp(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpexp(x), z); set_avma(av); }
 INLINE void mplogz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mplog(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mplog(x), z); set_avma(av); }
 INLINE void mpcosz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpcos(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpcos(x), z); set_avma(av); }
 INLINE void mpsinz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpsin(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpsin(x), z); set_avma(av); }
 INLINE void gnegz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(gneg(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(gneg(x), z); set_avma(av); }
 INLINE void gabsz(GEN x, long prec, GEN z)
-{ pari_sp av = avma; gaffect(gabs(x,prec), z); avma = av; }
+{ pari_sp av = avma; gaffect(gabs(x,prec), z); set_avma(av); }
 INLINE void gaddz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gadd(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gadd(x,y), z); set_avma(av); }
 INLINE void gsubz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gsub(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gsub(x,y), z); set_avma(av); }
 INLINE void gmulz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gmul(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmul(x,y), z); set_avma(av); }
 INLINE void gdivz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gdiv(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gdiv(x,y), z); set_avma(av); }
 INLINE void gdiventz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gdivent(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gdivent(x,y), z); set_avma(av); }
 INLINE void gmodz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gmod(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmod(x,y), z); set_avma(av); }
 INLINE void gmul2nz(GEN x, long s, GEN z)
-{ pari_sp av = avma; gaffect(gmul2n(x,s), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmul2n(x,s), z); set_avma(av); }
 INLINE void gshiftz(GEN x, long s, GEN z)
-{ pari_sp av = avma; gaffect(gshift(x,s), z); avma = av; }
+{ pari_sp av = avma; gaffect(gshift(x,s), z); set_avma(av); }
 
 /*******************************************************************/
 /*                                                                 */
@@ -2607,7 +2736,7 @@ INLINE long pr_get_e(GEN pr) { return gel(pr,3)[2]; }
 INLINE long pr_get_f(GEN pr) { return gel(pr,4)[2]; }
 INLINE GEN pr_get_tau(GEN pr){ return gel(pr,5); }
 INLINE int
-pr_is_inert(GEN P) { return pr_get_f(P) == lg(pr_get_gen(P))-1; }
+pr_is_inert(GEN P) { return typ(pr_get_tau(P)) == t_INT; }
 INLINE GEN
 pr_norm(GEN pr) { return powiu(pr_get_p(pr), pr_get_f(pr)); }
 INLINE ulong
@@ -2677,6 +2806,8 @@ nf_get_sign(GEN nf, long *r1, long *r2)
   *r2 = itou(gel(x,2));
 }
 
+INLINE GEN
+cyc_get_expo(GEN c) { return lg(c) == 1? gen_1: gel(c,1); }
 INLINE GEN
 abgrp_get_no(GEN x) { return gel(x,1); }
 INLINE GEN
@@ -2698,17 +2829,23 @@ bnf_get_reg(GEN bnf) { return gmael(bnf,8,2); }
 INLINE GEN
 bnf_get_logfu(GEN bnf) { return gel(bnf,3); }
 INLINE GEN
+bnf_get_sunits(GEN bnf)
+{ GEN s = gmael(bnf,8,3); return typ(s) == t_INT? NULL: s; }
+INLINE GEN
 bnf_get_tuU(GEN bnf) { return gmael3(bnf,8,4,2); }
 INLINE long
 bnf_get_tuN(GEN bnf) { return gmael3(bnf,8,4,1)[2]; }
 INLINE GEN
+bnf_get_fu_nocheck(GEN bnf) { return gmael(bnf,8,5); }
+INLINE GEN
 bnf_get_fu(GEN bnf) {
-  GEN fu = bnf_get_fu_nocheck(bnf);
+  GEN fu = bnf_build_units(bnf), nf = bnf_get_nf(bnf);
+  long i, l;
   if (typ(fu) == t_MAT) pari_err(e_MISC,"missing units in bnf");
+  l = lg(fu)-1; fu = vecslice(fu, 2, l);
+  for (i = 1; i < l; i++) gel(fu,i) = nf_to_scalar_or_alg(nf, gel(fu,i));
   return fu;
 }
-INLINE GEN
-bnf_get_fu_nocheck(GEN bnf) { return gmael(bnf,8,5); }
 
 INLINE GEN
 bnr_get_bnf(GEN bnr) { return gel(bnr,1); }
@@ -2833,6 +2970,8 @@ INLINE GEN
 rnf_get_disc(GEN rnf) { return gel(rnf,3); }
 INLINE GEN
 rnf_get_index(GEN rnf) { return gel(rnf,4); }
+INLINE GEN
+rnf_get_ramified_primes(GEN rnf) { return gel(rnf,5); }
 INLINE long
 rnf_get_varn(GEN rnf) { return varn(gel(rnf,1)); }
 INLINE GEN
@@ -2852,15 +2991,14 @@ INLINE GEN
 idealpseudored(GEN I, GEN G)
 { return ZM_mul(I, ZM_lll(ZM_mul(G, I), 0.99, LLL_IM)); }
 
-/* I integral (not necessarily HNF), G ZM, rounded Cholesky form of a weighted
- * T2 matrix. Return m in I with T2(m) small */
+/* Same I, G; m in I with T2(m) small */
 INLINE GEN
 idealpseudomin(GEN I, GEN G)
 {
   GEN u = ZM_lll(ZM_mul(G, I), 0.99, LLL_IM);
   return ZM_ZC_mul(I, gel(u,1));
 }
-/* I, G as in idealpseudomin. Return an irrational m in I with T2(m) small */
+/* Same I,G; irrational m in I with T2(m) small */
 INLINE GEN
 idealpseudomin_nonscalar(GEN I, GEN G)
 {
@@ -2869,6 +3007,26 @@ idealpseudomin_nonscalar(GEN I, GEN G)
   if (ZV_isscalar(m) && lg(u) > 2) m = ZM_ZC_mul(I, gel(u,2));
   return m;
 }
+/* Same I,G; t_VEC of irrational m in I with T2(m) small */
+INLINE GEN
+idealpseudominvec(GEN I, GEN G)
+{
+  long i, j, k, n = lg(I)-1;
+  GEN x, L, b = idealpseudored(I, G);
+  L = cgetg(1 + (n*(n+1))/2, t_VEC);
+  for (i = k = 1; i <= n; i++)
+  {
+    x = gel(b,i);
+    if (!ZV_isscalar(x)) gel(L,k++) = x;
+  }
+  for (i = 2; i <= n; i++)
+    for (j = 1; j < i; j++)
+    {
+      x = ZC_add(gel(b,i),gel(b,j));
+      if (!ZV_isscalar(x)) gel(L,k++) = x;
+    }
+  setlg(L,k); return L;
+}
 
 INLINE GEN
 idealred_elt(GEN nf, GEN I) {
